/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "imagePng.h"

 ImagePng::ImagePng(string filename, int color){
	 
	size_t size = filename.size() + 1;
    char * buffer = new char[ size ];
    strncpy( buffer, filename.c_str(), size ); 

	FILE *fp = NULL;	
	fp = fopen (buffer, "rb");
	if(!fp) {
		std::cerr << "error: couldn't open " << filename << std::endl;
    }

	/* Read magic number */
	png_byte magic[8];
	fread (magic, 1, sizeof (magic), fp);

  	/* Check for valid magic number */
	if (!png_check_sig (magic, sizeof (magic))) {
		std::cerr << "error: " << filename << " is not a valid PNG image!" << std::endl;
    	fclose (fp);
    }
	
	/* Create a png read struct */
  	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) {
		fclose (fp);
    }

	/* Create a png info struct */
	png_infop info_ptr = png_create_info_struct (png_ptr);
	if (!info_ptr) {
		fclose (fp);
		png_destroy_read_struct (&png_ptr, NULL, NULL);
	}

	if (setjmp (png_jmpbuf (png_ptr))) {
    	fclose (fp);
		png_destroy_read_struct (&png_ptr, &info_ptr, NULL);
	}
	
	/* Setup libpng for using standard C fread() function with our FILE pointer */
	png_init_io (png_ptr, fp);

	/* Tell libpng that we have already read the magic number */
	png_set_sig_bytes (png_ptr, sizeof (magic));

	/* Read png info */
	png_read_info (png_ptr, info_ptr);

	/* Get some usefull information from header */
	int bit_depth, color_type;
	bit_depth = png_get_bit_depth (png_ptr, info_ptr);
	color_type = png_get_color_type (png_ptr, info_ptr);
	
	/* Convert index color images to RGB images */
	if (color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb (png_ptr);

	/* Convert 1-2-4 bits grayscale images to 8 bits grayscale. */
	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		png_set_gray_1_2_4_to_8 (png_ptr);
	
	/* Convert grayscale to RBG */
	if(color) {
		if(color_type == PNG_COLOR_TYPE_GRAY) {
			color_type = PNG_COLOR_TYPE_RGB;
			png_set_gray_to_rgb(png_ptr);
		} else {
			if(color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
				color_type = PNG_COLOR_TYPE_RGBA;
				png_set_gray_to_rgb(png_ptr);
			}
		}
	}
	
	if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha (png_ptr);

	if (bit_depth == 16)
		png_set_strip_16 (png_ptr);
	else if (bit_depth < 8)
		png_set_packing (png_ptr);	
	
	/* Update info structure to apply transformations */
	png_read_update_info (png_ptr, info_ptr);

	/* Retrieve updated information */
	png_uint_32 w, h;
	png_get_IHDR (png_ptr, info_ptr, &w, &h, &bit_depth,&color_type, NULL, NULL, NULL);
	mWidth = w;
	mHeight = h;

	/* Get image format and components per pixel */
	setFormat(color_type);	
	  

	/* We can now allocate memory for storing pixel data */
	mTexels = new GLubyte[ mWidth * mHeight * internalFormat];	
	 
	/* Setup a pointer array.  Each one points at the begening of a row. */
	png_bytep *row_pointers = new png_bytep[mHeight];
	for(int i=0; i<mHeight; ++i) {
		row_pointers[i] = (png_bytep)(mTexels + ((mHeight-(i+1))*mWidth*internalFormat));
	}
	
	/* Read pixel data using row pointers */
	png_read_image (png_ptr, row_pointers);

	/* Finish decompression and release memory */
	png_read_end (png_ptr, NULL);
	png_destroy_read_struct (&png_ptr, &info_ptr, NULL);

	/* We don't need row pointers anymore */
	delete [] row_pointers;
	
	if(color) {
		Color coeffColor = Color(color, 0., 1.);
	
		for(int i=0;i<mHeight*mWidth*4;i+=4) {
			mTexels[i] *= coeffColor.r;
			mTexels[i+1] *= coeffColor.g;
			mTexels[i+2] *= coeffColor.b;
		}
	}

	fclose (fp);

}

ImagePng::ImagePng(int width, int height, GLenum format, GLubyte* texels):
	mWidth(width), mHeight(height)
{
	setFormatBis(format);	
	mTexels = texels;
}

ImagePng::ImagePng(wstring text, int color, int width, int height):
	mWidth(width), mHeight(height)
{
	setFormatBis(GL_RGBA);	
	ImagePng *imageChar;
	int widthChar, deltax, deltay;	

	int lenght = 0;

	for(int i=0;i<text.size();i++) {
		wchar_t c = text[i];
		getCharParam(c, widthChar, deltax, deltay);
		lenght+=widthChar;
	}
	
	if((text.size()==0)||(height<64)||(width<lenght)) {
		cout << "Empty text or too small text area \n";
		mTexels = NULL;
	} else {	
		mTexels = new GLubyte[mWidth * mHeight * 4];
		int index = 0;
		for(int i=0;i<mHeight;i++) {
			for(int j=0;j<mWidth;j++) {
				mTexels[index] = 0;
				mTexels[index+1] = 0;
				mTexels[index+2] = 0;
				mTexels[index+3] = 0;
				index+=4;
			}
		}

		int horSpace = (mWidth-lenght)/2;
		int vertSpace = mHeight/2-12;
		int horPos = horSpace;
		int vertPos;
	
		for(int k=0;k<text.size();k++) {
			wchar_t c = text[k];

			if(c!=L' ') {
				string im = getFilename (c);
				imageChar = new ImagePng(im,color);
				//delete im;
				getCharParam(c, widthChar, deltax, deltay);
				horPos-=deltax;
				vertPos = vertSpace -deltay;
				
				for(int i=0;i<imageChar->mHeight;i++) {
					for(int j=0;j<imageChar->mWidth;j++) {							
						int index = 4*((vertPos+i)*width+horPos+j);
						int indexBis = 4*(i*imageChar->mWidth+j);
						if(imageChar->mTexels[indexBis+3]>mTexels[index+3]) {
							mTexels[index] = imageChar->mTexels[indexBis];
							mTexels[index+1] = imageChar->mTexels[indexBis+1];
							mTexels[index+2] = imageChar->mTexels[indexBis+2];
							mTexels[index+3] = imageChar->mTexels[indexBis+3];
						}
					}
				}					
				horPos+=widthChar;
				delete imageChar;							
			} else {
				getCharParam(c, widthChar, deltax, deltay);					
				horPos+=widthChar;					
			}
		}
	}	
}

ImagePng::ImagePng(wstring text, int textColor, float textCoeff, int backgroundColor, float backgroundCoeff, int width, int height):
	mWidth(width), mHeight(height)
{
	setFormatBis(GL_RGBA);	
	ImagePng *imageChar;
	int widthChar, deltax, deltay;	

	int lenght = 0;

	for(int i=0;i<text.size();i++) {
		wchar_t c = text[i];
		getCharParam(c, widthChar, deltax, deltay);
		lenght+=widthChar;
	}
	
	if((text.size()==0)||(height<64)||(width<lenght)) {
		cout << "Empty text or too small text area \n";
		mTexels = NULL;
	} else {	
		Color background = Color(backgroundColor, backgroundCoeff, 1.);
		
		mTexels = new GLubyte[mWidth * mHeight * 4];
		
		int index = 0;
		for(int i=0;i<mHeight;i++) {
			for(int j=0;j<mWidth;j++) {
				mTexels[index] = static_cast<GLubyte>(background.r * 255.f);
				mTexels[index+1] = static_cast<GLubyte>(background.g * 255.f);
				mTexels[index+2] = static_cast<GLubyte>(background.b * 255.f);
				mTexels[index+3] = 0.;
				index+=4;
			}
		}

		int horSpace = (mWidth-lenght)/2;
		int vertSpace = mHeight/2-12;
		int horPos = horSpace;
		int vertPos;
	
		for(int k=0;k<text.size();k++) {
			wchar_t c = text[k];

			if(c!=L' ') {
				string im = getFilename (c);
				imageChar = new ImagePng(im,textColor);
				getCharParam(c, widthChar, deltax, deltay);
				horPos-=deltax;
				vertPos = vertSpace -deltay;
				
				for(int i=0;i<imageChar->mHeight;i++) {
					for(int j=0;j<imageChar->mWidth;j++) {							
						int index = 4*((vertPos+i)*width+horPos+j);
						int indexBis = 4*(i*imageChar->mWidth+j);
						if(imageChar->mTexels[indexBis+3] > mTexels[index+3]) {
							
							float alpha = imageChar->mTexels[indexBis+3] / 255.f;
							
							mTexels[index] = (1.-alpha) * mTexels[index] + alpha * imageChar->mTexels[indexBis];
							mTexels[index+1] = (1.-alpha) * mTexels[index+1] + alpha * imageChar->mTexels[indexBis+1];
							mTexels[index+2] = (1.-alpha) * mTexels[index+2] + alpha * imageChar->mTexels[indexBis+2];
							mTexels[index+3] = imageChar->mTexels[indexBis+3];
							
						} 
					}
				}					
				horPos+=widthChar;
				delete imageChar;							
			} else {
				getCharParam(c, widthChar, deltax, deltay);					
				horPos+=widthChar;					
			}
		}
		
		index = 3;
		for(int i=0;i<mHeight;i++) {
			for(int j=0;j<mWidth;j++) {
				mTexels[index] = 1.;
				index+=4;
			}
		}		
		
	}	

	
}


ImagePng::~ImagePng() {
	delete [] mTexels;
}


void ImagePng::setFormat(int colorType) {
	switch (colorType) {
		case PNG_COLOR_TYPE_GRAY:
 			mFormat = GL_LUMINANCE;
			internalFormat = 1;
			break;

		case PNG_COLOR_TYPE_GRAY_ALPHA:
			mFormat = GL_LUMINANCE_ALPHA;
			internalFormat = 2;
			break;

		case PNG_COLOR_TYPE_RGB:
			mFormat = GL_RGB;
			internalFormat = 3;
			break;

		case PNG_COLOR_TYPE_RGB_ALPHA:
			mFormat = GL_RGBA;
			internalFormat = 4;
			break;

		default:
			break;
    }
}

void ImagePng::setFormatBis(GLenum format) {
	mFormat = format;
	switch (format) {
		case GL_LUMINANCE:
			internalFormat = 1;
			break;

		case GL_LUMINANCE_ALPHA:
			internalFormat = 2;
			break;

		case GL_RGB:
			internalFormat = 3;
			break;

		case GL_RGBA:
			internalFormat = 4;
			break;

		default:
			break;
    }	
}

int ImagePng::getPngFormat() {
	int pngFormat;
	
	switch(internalFormat) {
		
		case 1:
			pngFormat = PNG_COLOR_TYPE_GRAY;
			break;

		case 2:
			pngFormat = PNG_COLOR_TYPE_GRAY_ALPHA;
			break;

		case 3:
			pngFormat = PNG_COLOR_TYPE_RGB;
			break;

		case 4:
			pngFormat = PNG_COLOR_TYPE_RGB_ALPHA;
			break;

		default:
			break;		
	}
	
	return pngFormat;
}


void ImagePng::setTexels(GLubyte* texels) {
	mTexels = texels;
}
	

GLuint ImagePng::getWidth() {
	return mWidth;
}

GLuint ImagePng::getHeight() {
	return mHeight;	
}

GLenum ImagePng::getFormat() {
	return mFormat;
}

int ImagePng::getInternalFormat() {
	return internalFormat;
}

GLubyte* ImagePng::getTexels() {
	return mTexels;
}


void ImagePng::raster2D(int wPosition, int hPosition) {
	glRasterPos2i(wPosition,hPosition);	
	glDrawPixels(mWidth, mHeight, mFormat, GL_UNSIGNED_BYTE, mTexels);	
}


bool ImagePng::writeToFile(string filename) {
	FILE *fp = fopen(filename.c_str(), "wb");
	
	if (!fp) {
		return false;
	}

	//png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, (png_voidp)user_error_ptr, user_error_fn, user_warning_fn);
	png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) {
		fclose(fp);		
		return false;
	}

	png_infop info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
		fclose(fp);		
		return false;
    }
	
	if (setjmp(png_jmpbuf(png_ptr))) {
		png_destroy_write_struct(&png_ptr, &info_ptr);
		fclose(fp);
		return false;
	}

    png_init_io(png_ptr, fp);
	
	png_set_IHDR(png_ptr, info_ptr, mWidth, mHeight, 8, getPngFormat(), PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	
	png_write_info(png_ptr, info_ptr);
	
	for (int i=0; i<mHeight; i++) {
		png_write_row(png_ptr, (png_bytep)(mTexels + ((mHeight-(i+1))*mWidth*internalFormat)));
	}
	
	png_write_end(png_ptr, NULL);	
	
	png_destroy_write_struct(&png_ptr, &info_ptr);
	fclose(fp);	
	
	return true;
}
